/*==========================================================================
	Datei:	mem_alloc.c

	Implementation Malloc-berwachung

	Begonnen:			22.08.2010
	Letzte nderung:	22.08.2010
	Autor:				M.Bhme\HAMEG Chemnitz
===========================================================================*/

/*--- Includes ------------------------------------------------------------*/
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <assert.h>

#include <mem_alloc.h>

/*--- Makros & Konstanten -------------------------------------------------*/

// malloc braucht wie snprintf ebenfalls pro IRQ-Level eine reent-Struktur

#if !defined HW_MC_MAIN_CPU_TYPE || HW_MC_MAIN_CPU_TYPE == HW_MC_CPU_TYPE_INTEL_I686
// im Emulator gibt's die CPU-Funktionen nicht
#define MEM_ALLOC_DISABLE_IRQ
#define MEM_ALLOC_ENABLE_IRQ

#else
#if HW_MC_MAIN_CPU_TYPE == HW_MC_CPU_TYPE_ADSP_BF561
// !!! Da nicht jeder Interruptlevel (== Thread) seinen eigenen Heap hat, darf das reent nicht verwendet werden
// -> man muss whrend der Allokation eigentlich nur die Interruptlevel sperren, die ebenfalls malloc verwenden
// IRQ's werden ggf. im __malloc_lock / __malloc_unlock verboten/erlaubt
#define MEM_ALLOC_USE_REENT
#endif

#if 0	// IRQ's werden ggf. im __malloc_lock / __malloc_unlock verboten/erlaubt
#define MEM_ALLOC_DISABLE_IRQ		CPU_disableIRQ0();
#define MEM_ALLOC_ENABLE_IRQ		CPU_enableIRQ0();
#else
#define MEM_ALLOC_DISABLE_IRQ
#define MEM_ALLOC_ENABLE_IRQ
#endif

#endif

/*--- Type Definitionen ---------------------------------------------------*/

/*--- Prototypen ----------------------------------------------------------*/

/*--- Globals -------------------------------------------------------------*/

#if !defined NDEBUG || defined MEM_ALLOC_DEBUG || defined MEM_ALLOC_DEBUG_TOTAL
// Gesamtstatistik
struct mem_allocModuleInfo	mem_allocTotal;
#endif

#ifdef MEM_ALLOC_DEBUG
#warning "Allokierter Speicher wird mitgezaehlt!"

// berwachung der einzelnen Module
struct mem_allocModuleInfo	mem_allocModules[MEM_ALLOC_MODULE_NO];


#ifdef MEM_ALLOC_DEBUG_FENCE
#warning "Allokierter Speicher wird auf Pufferueberlaeufe getestet! Das bremst etwas..."

const struct mem_allocTableFence	mem_fencePatternPre  = {{0xde, 0xad, 0xaa, 0xaa}};
const struct mem_allocTableFence	mem_fencePatternPost = {{0xde, 0xad, 0xee, 0xee}};

int32_t						mem_fenceTotalMem = 0;

// vor dem allokierten Speicher 4 Byte Platz lassen und mit dem Muster fllen
#define MEM_ALLOC_PRE_FENCE

// nach dem allokierten Speicher 4 Byte Platz lassen und mit dem Muster fllen
#define MEM_ALLOC_POST_FENCE

#endif	// MEM_ALLOC_DEBUG_FENCE

#ifdef MEM_ALLOC_LOG_MODULE_ID
// Allokierungsvorgnge eines bestimmten Moduls berwachen
#warning "Allokierter Speicher eines Moduls wird tabellerisch ueberwacht! Das bremst etwas mehr..."

#ifdef MEM_ALLOC_LOG_IGNORE_MODULE_ID
#define MEM_ALLOC_LOG_TEST_MODULE_ID(moduleId)		(MEM_ALLOC_LOG_IGNORE_MODULE_ID != moduleId && (MEM_ALLOC_LOG_MODULE_ID == MEM_ALLOC_MODULE_NO || moduleId == MEM_ALLOC_LOG_MODULE_ID))
#else
#define MEM_ALLOC_LOG_TEST_MODULE_ID(moduleId)		(MEM_ALLOC_LOG_MODULE_ID == MEM_ALLOC_MODULE_NO || moduleId == MEM_ALLOC_LOG_MODULE_ID)
#endif

// Anzahl der Eintrge in der berwachungstabelle
#ifndef MEM_ALLOC_LOG_TABLE_SIZE
#define MEM_ALLOC_LOG_TABLE_SIZE		512
#endif

#define MEM_ALLOC_TABLE_ENTRY_USED		(1<<0)


// Tabelle aller allokierten Speicherblcke
struct mem_allocTableEntry	mem_allocTable[MEM_ALLOC_LOG_TABLE_SIZE];
const uint32_t				mem_allocTableSize = MEM_ALLOC_LOG_TABLE_SIZE;

struct mem_allocTableEntry	*mem_allocTableUsedHead;
struct mem_allocTableEntry	*mem_allocTableUsedTail;
struct mem_allocTableEntry	*mem_allocTableFreeHead;
struct mem_allocTableEntry	*mem_allocTableFreeTail;

// Index des zuletzt eingetragenen Speicherblocks
int32_t						mem_allocTableAddIndex = 0;
// Index des zuletzt ausgetragenen Speicherblocks
int32_t						mem_allocTableRemoveIndex = 0;


/*--- Funktionen ----------------------------------------------------------*/

void mem_allocTableInit (void)
{
	int32_t	id;

	// Liste verketten
	for (id = 0; id < (MEM_ALLOC_LOG_TABLE_SIZE - 1); id++)
		mem_allocTable[id].next = &mem_allocTable[id + 1];

	// Letzter Eintrag
	mem_allocTable[MEM_ALLOC_LOG_TABLE_SIZE - 1].next = NULL;

	// Listenzeiger
	mem_allocTableFreeHead = &mem_allocTable[0];
	mem_allocTableFreeTail = &mem_allocTable[MEM_ALLOC_LOG_TABLE_SIZE - 1];
	mem_allocTableUsedHead = NULL;
	mem_allocTableUsedTail = NULL;
}


// Eintragen
int32_t mem_addAllocTableEntry (const uint8_t * start, size_t size, unsigned int moduleId)
{
#if 1
	struct mem_allocTableEntry	*entry;

	// Kopf der freien Liste herauslsen
	entry = mem_allocTableFreeHead;
	if (entry == NULL) {
		if (mem_allocTableUsedHead == NULL) {
			// Tabelle noch nicht initialisiert
			mem_allocTableInit ();

			// ersten freien Eintrag holen
			entry = mem_allocTableFreeHead;
		} else {
			// nichts mehr frei
			return -1;
		}
	}
	// Nachfolger wird neuer Kopf
	mem_allocTableFreeHead = entry->next;
	// wenn das das letzte freie Element war -> Ende auch auf NULL
	if (mem_allocTableFreeHead == NULL)
		mem_allocTableFreeTail = NULL;

	// Eintrag wird ans Ende der Liste der belegten angehngt
	entry->next = NULL;
	if (mem_allocTableUsedTail != NULL) {
		// als Nachfolger ans Ende
		mem_allocTableUsedTail->next = entry;
		mem_allocTableUsedTail = entry;
	} else {
		// allererster Eintrag
		mem_allocTableUsedHead = entry;
		mem_allocTableUsedTail = entry;
	}

	// Eintrag fllen
	entry->mm = start;
	entry->sz = size;
	entry->id = (enum MEM_ALLOC_MODULE_LIST)moduleId;

#ifdef MEM_ALLOC_DEBUG_FENCE
	// zustzliche Zeiger auf die Gartenzune
#ifdef MEM_ALLOC_PRE_FENCE
	entry->he = (struct mem_allocTableFence *)(start - sizeof(struct mem_allocTableFence));
#endif
#ifdef MEM_ALLOC_POST_FENCE
	entry->ta = (struct mem_allocTableFence *)(start + size);
#endif
#endif

	return 0;

#else
	int32_t	id;

	// Tabelle durchsuchen nach einem freien Eintrag
	// beginnend ab dem zuletzt eingefgten Eintrag
	id = mem_allocTableAddIndex;
	do {
		// Index erstmal eins weiter und prfen auf berlauf
		if (++id >= MEM_ALLOC_LOG_TABLE_SIZE)
			id = 0;
		// Eintrag prfen
		if ((mem_allocTable[id].st & MEM_ALLOC_TABLE_ENTRY_USED) == 0) {
			// Eintrag belegen
			mem_allocTable[id].st |= MEM_ALLOC_TABLE_ENTRY_USED;
			mem_allocTable[id].mm = start;
			mem_allocTable[id].sz = size;
			mem_allocTable[id].id = (enum MEM_ALLOC_MODULE_LIST)moduleId;

#ifdef MEM_ALLOC_DEBUG_FENCE
			// zustzliche Zeiger auf die Gartenzune
#ifdef MEM_ALLOC_PRE_FENCE
			mem_allocTable[id].he = (struct mem_allocTableFence *)(start - sizeof(uint32_t));
#endif
#ifdef MEM_ALLOC_POST_FENCE
			mem_allocTable[id].ta = (struct mem_allocTableFence *)(start + size);
#endif
#endif
			// Index merken, die nchste Suche startet auf dem Eintrag dahinter
			mem_allocTableAddIndex = id;
			// und raus
			return id;
		}
		// solange bis der Startwert wieder erreicht wird
	} while (id != mem_allocTableAddIndex);

	// kein Eintrag frei
//	assert (start == NULL);
#endif
	return -1;
}

// Austragen
size_t mem_removeAllocTableEntry (const uint8_t * start, size_t size, unsigned int moduleId)
{
#if 1
	struct mem_allocTableEntry	*entry, *prev;

	prev = NULL;
	// Eintrag in der Liste der verwendeten suchen
	entry = mem_allocTableUsedHead;
	while (entry != NULL) {
		// Adresse prfen
		if (entry->mm == start) {
			// Eintrag testen
			if (size > 0 && entry->sz != size) {
				// das passt nicht zusammen
				assert (entry->sz == size);
			}
			// die ModulID muss nicht unbedingt bereinstimmen, aber es wre ein Zeichen fr sauberes Programmieren
//			assert (entry->moduleId == moduleId);

			// Eintrag freigeben
			if (entry == mem_allocTableUsedHead) {
				// Kopf entfernen
				mem_allocTableUsedHead = entry->next;
				// wenn das der letzte Eintrag war -> Ende auch auf NULL
				if (mem_allocTableUsedHead == NULL)
					mem_allocTableUsedTail = NULL;

			} else if (entry == mem_allocTableUsedTail) {
				// Ende entfernen
				prev->next = entry->next;
				mem_allocTableUsedTail = prev;

			} else {
				// mittendrin
				prev->next = entry->next;
			}

			// ans Ende der freien Liste anhngen
			entry->next = NULL;
			if (mem_allocTableFreeTail != NULL) {
				// ans Ende anhngen
				mem_allocTableFreeTail->next = entry;
				mem_allocTableFreeTail = entry;
			} else {
				// erster Eintrag
				mem_allocTableFreeHead = entry;
				mem_allocTableFreeTail = entry;
			}

			// Gre zurckgeben
			return entry->sz;
		}
		// Vorgnger merken
		prev = entry;
		// nchster Eintrag
		entry = entry->next;
	}

	// nicht gefunden
	return 0;

#else
	int32_t id;

	// Tabelle rckwrts durchsuchen nach dem zugehrigen Eintrag
	// beginnend ab dem zuletzt eingefgten Eintrag
	id = mem_allocTableAddIndex;
	do {
		// Index erstmal eins zurck und prfen auf berlauf
		if (--id < 0)
			id = MEM_ALLOC_LOG_TABLE_SIZE - 1;
		// Eintrag prfen
		if ((mem_allocTable[id].st & MEM_ALLOC_TABLE_ENTRY_USED) != 0 && mem_allocTable[id].mm == start) {
			// Eintrag testen
			if (size > 0 && mem_allocTable[id].sz != size) {
				// das passt nicht zusammen
				assert (mem_allocTable[id].sz == size);
			}
			// immer die eingetragene Gre des Speicherblocks zurckgeben, auf die kann man sich verlassen
			size = mem_allocTable[id].sz;
			// die ModulID muss nicht unbedingt bereinstimmen, aber es wre ein Zeichen fr sauberes Programmieren
//			assert (mem_allocTable[id].moduleId == moduleId);
			// nur der Status wird beim Eintragen geprft, der Rest kann seine Werte behalten
			mem_allocTable[id].st &= ~MEM_ALLOC_TABLE_ENTRY_USED;

			// merken fr's Debuggen
			mem_allocTableRemoveIndex = id;

			// und raus
			return size;
		}
		// solange bis der Startwert wieder erreicht wird
	} while (id != mem_allocTableAddIndex);

	// Eintrag nicht gefunden
//	assert (start == NULL);

	return 0;
#endif
}

#endif	// MEM_ALLOC_LOG_MODULE_ID
#endif	// MEM_ALLOC_DEBUG


void * mem_alloc (size_t size, unsigned int moduleId)
{
	// realloc macht das gleiche wie malloc, wenn man als Zeiger ne NULL bergibt
	return mem_realloc (NULL, 0, size, moduleId);
}

void * mem_calloc (size_t elems, size_t elemSize, unsigned int moduleId)
{
	void * tmp;
	size_t size;

	size = elems * elemSize;

	// realloc macht das gleiche wie malloc, wenn man als Zeiger ne NULL bergibt
	tmp = mem_realloc (NULL, 0, size, moduleId);
	// nach dem Allokieren den Speicher lschen
	if (tmp != NULL)
		memset (tmp, 0x00, size);
	// Ergebnis zurck
	return tmp;
}

void mem_free (void * memPtr, size_t size, unsigned int moduleId)
{
	// unser realloc macht das gleiche wie free, wenn man als neue Gre 0 bergibt, so wie es im Handbuch steht
	// in der libc muss dafr ein Schalter gesetzt sein (der aber im Blackfin nicht gesetzt ist!!!), daher rufen wir explizit free()
	mem_realloc (memPtr, size, 0, moduleId);
}

void * mem_realloc (void * memPtr, size_t oldSize, size_t newSize, unsigned int moduleId)
{
	uint8_t * tmp;
	size_t allocSize;

	tmp = (uint8_t *)memPtr;

	allocSize = 0;

#if !defined NDEBUG || defined MEM_ALLOC_DEBUG || defined MEM_ALLOC_DEBUG_TOTAL
	// Gesamtstatistik
	mem_allocTotal.active = 1;
#endif

#ifdef MEM_ALLOC_DEBUG
	// aktiv merken
	if (moduleId < MEM_ALLOC_MODULE_NO) {
		mem_allocModules[moduleId].active = 1;
	}
#endif

	// zu allokierender Speicher
	allocSize = newSize;

	// bergebenen Zeiger testen
	if (tmp == NULL) {
		// kein Zeiger bergeben
		if (newSize > 0) {
			// kein Zeiger bergeben, aber eine Gre fr neuen Speicher -> tut das gleiche wie malloc
#if !defined NDEBUG || defined MEM_ALLOC_DEBUG || defined MEM_ALLOC_DEBUG_TOTAL
			// gesamter allokierter Speicher
			mem_allocTotal.mallocCnt++;
#endif
#ifdef MEM_ALLOC_DEBUG
			// Modul-Statistik
			if (moduleId < MEM_ALLOC_MODULE_NO) {
				mem_allocModules[moduleId].mallocCnt++;
			}

			// zu allokierender Speicher
			allocSize = newSize;

#ifdef MEM_ALLOC_DEBUG_FENCE
			// zur Erkennung von Pufferberlufen vor und nach dem
			// allokierten Speicher etwas Platz lassen und mit einer
			// bekannten Zahl fllen
#ifdef MEM_ALLOC_PRE_FENCE
			allocSize += sizeof(struct mem_allocTableFence);
#endif
#ifdef MEM_ALLOC_POST_FENCE
			allocSize += sizeof(struct mem_allocTableFence);
#endif
#endif	// MEM_ALLOC_DEBUG_FENCE
#endif	// MEM_ALLOC_DEBUG
		} else {
			// kein Zeiger und neue Gre identisch 0 -> verhlt sich wie free(NULL)
#ifdef MEM_ALLOC_DEBUG
			// aktiv lschen
			if (moduleId < MEM_ALLOC_MODULE_NO) {
				mem_allocModules[moduleId].active = 0;
			}
#endif

#if !defined NDEBUG || defined MEM_ALLOC_DEBUG || defined MEM_ALLOC_DEBUG_TOTAL
			// Gesamtstatistik
			mem_allocTotal.active = 0;
#endif
			// sofort raus
			return NULL;
		}
	} else {
		// wenn der Zeiger schon wohin zeigt

		if (newSize == 0) {
			// Zeiger auf Speicher bergeben und neue Gre ist 0 -> tut das gleiche wie free
#if !defined NDEBUG || defined MEM_ALLOC_DEBUG || defined MEM_ALLOC_DEBUG_TOTAL
			// gesamter allokierter Speicher
			mem_allocTotal.freeCnt++;
#endif
#ifdef MEM_ALLOC_DEBUG
			// Modul-Statistik
			if (moduleId < MEM_ALLOC_MODULE_NO) {
				mem_allocModules[moduleId].freeCnt++;
			}
#endif
		} else {
			// Zeiger auf Speicher bergeben und neue Gre ist grer 0 -> echtes realloc

#if !defined NDEBUG || defined MEM_ALLOC_DEBUG || defined MEM_ALLOC_DEBUG_TOTAL
			// gesamter allokierter Speicher
			mem_allocTotal.reallocCnt++;
#endif
#ifdef MEM_ALLOC_DEBUG
			// Modul-Statistik
			if (moduleId < MEM_ALLOC_MODULE_NO) {
				mem_allocModules[moduleId].reallocCnt++;
			}
#endif

			// erneut zu allokierender Speicher
			allocSize = newSize;

#ifdef MEM_ALLOC_DEBUG_FENCE
			// zur Erkennung von Pufferberlufen vor und nach dem
			// allokierten Speicher etwas Platz lassen und mit einer
			// bekannten Zahl fllen
#ifdef MEM_ALLOC_PRE_FENCE
			allocSize += sizeof(struct mem_allocTableFence);
#endif
#ifdef MEM_ALLOC_POST_FENCE
			allocSize += sizeof(struct mem_allocTableFence);
#endif
#endif	// MEM_ALLOC_DEBUG_FENCE
		}

#ifdef MEM_ALLOC_LOG_MODULE_ID
		// Speicherblock aus der Tabelle austragen
		if (MEM_ALLOC_LOG_TEST_MODULE_ID(moduleId)) {
			// liefert die beim Allokieren angegebene Gre zurck
			size_t tabSize = mem_removeAllocTableEntry (tmp, oldSize, moduleId);

			// nimm die in der Tabelle gefundene Gre, falls der Rufer nichts bergeben hat
			if (oldSize == 0 && tabSize > 0)
				oldSize = tabSize;
		}
#endif

#ifdef MEM_ALLOC_DEBUG_FENCE
#ifdef MEM_ALLOC_POST_FENCE
		// am Ende des allokierten Speichers 4 Byte auf die bekannte Zahlenfolge prfen
		if (oldSize > 0) {
			// geht nur wenn die Gre bekannt ist
			// prfen, ob nach dem Ende geschrieben wurde
			if (memcmp (tmp + oldSize, &mem_fencePatternPost, sizeof(struct mem_allocTableFence)) != 0) {
				// passt nicht !
				assert (0 /* Speicher direkt nach dem allokierten Bereich berschrieben! */);
			}
		}
		// ebenfalls merken, wieviel Speicher inklusive dem Gartenzaun allokiert wurde
		mem_fenceTotalMem -= sizeof(struct mem_allocTableFence);
#endif

#ifdef MEM_ALLOC_PRE_FENCE
		// wenn davor ein uint32_t eingefgt wurde, muss der Zeiger wieder zurckgeschoben werden
		tmp -= sizeof(struct mem_allocTableFence);
		// und prfen, ob vor dem Start geschrieben wurde
		if (memcmp (tmp, &mem_fencePatternPre, sizeof(struct mem_allocTableFence)) != 0) {
			// passt nicht !
			assert (0 /* Speicher direkt vor dem allokierten Bereich berschrieben! */);
		}
		// ebenfalls merken, wieviel Speicher inklusive dem Gartenzaun allokiert wurde
		mem_fenceTotalMem -= sizeof(struct mem_allocTableFence);
#endif
#endif	// MEM_ALLOC_DEBUG_FENCE

		// Ist die Gre des bergebenen Speicherblocks bekannt ?
		if (oldSize > 0) {
#if !defined NDEBUG || defined MEM_ALLOC_DEBUG || defined MEM_ALLOC_DEBUG_TOTAL
			// gesamter allokierter Speicher
			if (mem_allocTotal.totalMem >= oldSize)
				mem_allocTotal.totalMem -= oldSize;
			else
				assert (0 /* bergebene Gre falsch! */);
#endif
#ifdef MEM_ALLOC_DEBUG
			// Modul-Statistik
			if (moduleId < MEM_ALLOC_MODULE_NO) {
				if (mem_allocModules[moduleId].totalMem >= oldSize)
					mem_allocModules[moduleId].totalMem -= oldSize;
				else
					assert (0 /* bergebene Gre falsch! */);
			}
#endif
#ifdef MEM_ALLOC_DEBUG_FENCE
			// ebenfalls merken, wieviel Speicher inklusive dem Gartenzaun allokiert wurde
			if (mem_fenceTotalMem >= oldSize)
				mem_fenceTotalMem -= oldSize;
			else
				assert (0 /* bergebene Gre falsch! */);
#endif
		}
	}

	// Bibliotheksfunktionen rufen
	if (tmp == NULL) {
		// kein Zeiger bergeben
		if (allocSize > 0) {
			// malloc
#ifdef MEM_ALLOC_USE_REENT
			tmp = (unsigned char *)_malloc_r(__getreent(), allocSize);
#else
			// eine Allokation darf von keiner anderen Allokation unterbrochen werden
			MEM_ALLOC_DISABLE_IRQ
			tmp = (uint8_t *)malloc(allocSize);
			MEM_ALLOC_ENABLE_IRQ
#endif
		} else {
			// nix tun
			tmp = NULL;
		}
	} else {
		// Zeiger bergeben
		if (allocSize > 0) {
			// realloc
#ifdef MEM_ALLOC_USE_REENT
			tmp = (unsigned char *)_realloc_r(__getreent(), tmp, allocSize);
#else
			// eine Allokation darf von keiner anderen Allokation unterbrochen werden
			MEM_ALLOC_DISABLE_IRQ
			tmp = (uint8_t *)realloc(tmp, allocSize);
			MEM_ALLOC_ENABLE_IRQ
#endif
		} else {
			// free
#ifdef MEM_ALLOC_USE_REENT
			_free_r (__getreent(), tmp);
#else
			// eine Allokation darf von keiner anderen Allokation unterbrochen werden
			MEM_ALLOC_DISABLE_IRQ
			free (tmp);
			MEM_ALLOC_ENABLE_IRQ
#endif
			tmp = NULL;
		}
	}

	// Nur wenn wirklich neuer Speicher allokiert werden sollte und konnte !
	if (newSize > 0 && tmp != NULL) {
#if !defined NDEBUG || defined MEM_ALLOC_DEBUG || defined MEM_ALLOC_DEBUG_TOTAL
		// gesamter allokierter Speicher
		mem_allocTotal.totalMem += newSize;
#endif
#ifdef MEM_ALLOC_DEBUG
		// Modul-Statistik
		if (moduleId < MEM_ALLOC_MODULE_NO) {
			mem_allocModules[moduleId].totalMem += newSize;
		}

#ifdef MEM_ALLOC_DEBUG_FENCE
		// ebenfalls merken, wieviel Speicher inklusive dem Gartenzaun allokiert wurde
		mem_fenceTotalMem += allocSize;

#ifdef MEM_ALLOC_PRE_FENCE
		// die ersten 4 Byte vor dem Speicher mit der bekannten Zahlenfolge belegen
		memcpy (tmp, &mem_fencePatternPre, sizeof(struct mem_allocTableFence));
		// und den zurckgegebenen Zeiger um diese 4 Bytes weiterschieben
		tmp += sizeof(struct mem_allocTableFence);
#endif

#ifdef MEM_ALLOC_POST_FENCE
		// am Ende des allokierten Speichers 4 Byte mit der bekannten Zahlenfolge belegen
		memcpy (tmp + newSize, &mem_fencePatternPost, sizeof(struct mem_allocTableFence));
#endif
#endif	// MEM_ALLOC_DEBUG_FENCE

#ifdef MEM_ALLOC_LOG_MODULE_ID
		// neuen Speicherblock in die Tabelle eintragen
		if (MEM_ALLOC_LOG_TEST_MODULE_ID(moduleId)) {
			// nur die angeforderte Gre eintragen, die wird beim Austragen wieder gebraucht
			mem_addAllocTableEntry (tmp, newSize, moduleId);
		}
#endif
#endif	// MEM_ALLOC_DEBUG
	} else {
		if (newSize > 0) {
			// Nullzeiger -> kein Speicher erhalten!
			// im Debug anhalten
			assert (tmp != NULL);
		}
	}

#ifdef MEM_ALLOC_DEBUG
	// aktiv lschen
	if (moduleId < MEM_ALLOC_MODULE_NO) {
		mem_allocModules[moduleId].active = 0;
	}
#endif

#if !defined NDEBUG || defined MEM_ALLOC_DEBUG || defined MEM_ALLOC_DEBUG_TOTAL
	// Gesamtstatistik
	mem_allocTotal.active = 0;
#endif

	// Ergebnis zurck
	return (void *)tmp;
}
